Library IEEE;
use ieee.std_logic_1164.all;

use work.sp_defs.all;
use work.instructions.all;


-- note: gives error warning showing that signal "result" was sign extended
entity Instruction is
	port (	iin       : in InstructionCode;
			reset, nz, clock : in std_logic;
			iout      : out Opcode );

end;

architecture behaviour2 of Instruction is

	  signal instruction, opnext : InstructionCode;

	  constant nop   : Opcode := "000000000000000"; -- instruction register
	
	  constant aor   : Opcode := "001000000000000"; -- function unit
	  constant ntr   : Opcode := "010000000000000";
	  constant ltz   : Opcode := "011000000000000";
	  constant plus  : Opcode := "100000000000000";
	  constant minus : Opcode := "101000000000000";
	
	  constant pstay : Opcode := "000010000000000"; -- program counter
	  constant pstom : Opcode := "000100000000000";
	  constant pstor : Opcode := "000110000000000";
	
	  constant mrdp  : Opcode := "000000000000000"; -- memory
	  constant mrdt  : Opcode := "000000100000000";
	  constant mstop : Opcode := "000001000000000";
	  constant mstot : Opcode := "000001100000000";
	
	  constant tstof : Opcode := "000000010000000"; -- top register
	  constant tstor : Opcode := "000000011000000";
	
	  constant dpop  : Opcode := "000000000010000"; -- data stack
	  constant dpshu : Opcode := "000000000100000";
	  constant dpshm : Opcode := "000000000101000";
	  constant dstou : Opcode := "000000000110000";
	  constant dstom : Opcode := "000000000111000";
	
	  constant rpop  : Opcode := "000000000000010"; -- return stack
	  constant rpsht : Opcode := "000000000000100";
	  constant rpshp : Opcode := "000000000000101";
	  constant rstot : Opcode := "000000000000110";
	  constant rstop : Opcode := "000000000000111";
	
begin

  process(instruction,nz) begin  			-- ouput decode
    case instruction is

		-- reset nop
		when ext_nop =>
			iout   <= nop;
			opnext <= int_nop;

		-- internal nop, no memory access
      	when int_nop => 
			iout   <= nop;  			
           	opnext <= load_mem;

		-- memory contents to DS (can be used for a "store immediate" instruction)
      	when sto_mem_DS => 	
			iout   <= mrdp or dstom;
			opnext <= int_nop;

      	when psh_mem_DS => 
			iout   <= mrdp or dpshm;
			opnext <= int_nop;

      	when pop_DS_TOP => 
			iout   <= dpop or tstof;
			opnext <= load_mem;

      	when sto_DS_TOP => 
			iout   <= tstof;
			opnext <= load_mem;

		when swap =>
			iout   <= tstof or dstou;
			opnext <= load_mem;

		when sto_TOP_DS =>
			iout   <= dstou;
			opnext <= load_mem;

		-- store the data stack to the given memory location
      	when sto_DS_memimm => 
			iout   <= mrdp or pstom or rstop;
			opnext <= sto_DS_memimm2;

      	when sto_DS_memimm2 => 
			iout   <= mstop or pstor;
			opnext <= ext_nop;

      	when sto_memimm_DS_and_sto_DS_TOP => 
			iout   <= tstof or pstom or rstop;
			opnext <= sto_memimm_DS2;

		when sto_memimm_DS2 =>
			iout   <= pstor;
			opnext <= sto_memimm_DS3;

		when sto_memimm_DS3 =>
			iout   <= dstom;
			opnext <= int_nop;

		when psh_memptr_DS =>
			iout <= pstom or rstop;
			opnext <= psh_memptr_DS2;
		when psh_memptr_DS2 =>
			iout <= pstay;
			opnext <= psh_memptr_DS3;
		when psh_memptr_DS3 =>
			iout <= pstom;
			opnext <= psh_memptr_DS4;
		when psh_memptr_DS4 =>
			iout <= pstor;
			opnext <= psh_memptr_DS5;
		when psh_memptr_DS5 =>
			iout <= dpshm;
			opnext <= int_nop;

		-- branch if not zero to next memory location
      	when bnzero_imm => 
			if nz = '1' then	-- branch
				iout   <= pstom;
				opnext <= ext_nop;
			else
				iout   <= nop;  -- skip over next mem. loc.
				opnext <= int_nop;
			end if;

		-- branch to given memory location
      	when bra_mem => 
			iout   <= pstom;
			opnext <= ext_nop;

		-- branch if zero to next memory location (reverse of bnzero)
      	when bzero_imm => 
			if nz = '0' then	-- branch
				iout   <= pstom;
				opnext <= ext_nop;
			else
				iout   <= nop;  -- skip over next mem. loc.
				opnext <= int_nop;
			end if;

		-- add (NOTE: there has to be some "filler" in the program memory after the
		--		      add instruction in order for this instr. to work correctly. 
		--			  (just think of an add instruction taking up 2 locations instead
		--				of one.) )
      	when add => 
			if nz = '1' then
				iout   <= pstay or plus or tstof or dstou;
				opnext <= add;
			else
				iout   <= nop;
				opnext <= int_nop;
			end if;

		-- subtract ( note for "add" applies here. )
      	when subtract => 
			if nz = '1' then
				iout   <= pstay or minus or tstof or dstou;
				opnext <= add;  -- rest of operations are adds
			else
				iout   <= nop;
				opnext <= int_nop;
			end if;

		when incr_ptr =>
			iout <= dpshm or pstay;				-- store the pointer to DS
			opnext <= incr_ptr2;
		when incr_ptr2 =>
			iout <= tstof or pstay;				-- store DS -> top
			opnext <= incr_ptr3;
		when incr_ptr3 =>
			iout <= mrdt or pstay;
			opnext <= incr_ptr4;
		when incr_ptr4 =>
			iout <= pstom or rstop;				-- swap RS and PC
			opnext <= incr_ptr9;
		when incr_ptr9 =>
			iout <= nop;						-- do a nop to increment PC.
			opnext <= incr_ptr5;
		when incr_ptr5 =>
			iout <= pstor or rstop;				-- swap RS and PC
			opnext <= incr_ptr6;
		when incr_ptr6 =>
			iout <= pstay or tstor;				-- TOP stores RS
			opnext <= incr_ptr7;
		when incr_ptr7 =>
			iout <= pstay or tstof or dstou;  	-- swap TOP and DS
			opnext <= incr_ptr8;
		when incr_ptr8 =>
			iout <= mstot or pstay or dpop;		-- store DS -> mem(TOP)
			opnext <= ext_nop;
	

		-- if a stop instruction is issued, this takes care of it...
		when others => 	-- discontinue processing;
				iout   <= pstay;			 	
                opnext <= stop;

    end case;
  end process;
  
  process(clock,reset)begin  -- clock in opcode
    	if  reset = '1'  then
      		instruction <= ext_nop;
    	elsif  clock = '1' and clock'event  then
    	  	if  opnext = load_mem then
        		instruction <= iin; -- fetch opcode from memory
      		else
        		instruction <= opnext; -- use internal opcode
      		end if;
    	end if;
  end process;
end behaviour2;
