library IEEE;
use IEEE.STD_Logic_1164.all;

package stack_processor is
	component stack_processor
		generic( data_stack_index_width : integer := 3; 
			  data_stack_index_depth : integer := 8; 
			  return_stack_index_width : integer := 2; 
			  return_stack_index_depth : integer := 4 
			);
		port (
			reset, clock : in std_logic;
			memdata_in : in std_logic_vector( 7 downto 0);
			memdata_out : out std_logic_vector ( 7 downto 0);
			address_out : out std_logic_vector( 7 downto 0);
			fetch_store : out std_logic;
			topout : out std_logic_vector( 7 downto 0);
			pcout : out std_logic_vector( 7 downto 0)
		);
	end component;
end stack_processor;

library IEEE;
use IEEE.STD_Logic_1164.all;

use work.sp_defs.all;

entity stack_processor is
	generic ( data_stack_index_width : integer := 3; 
			  data_stack_index_depth : integer := 8; 
			  return_stack_index_width : integer := 2; 
			  return_stack_index_depth : integer := 4 
			);
	port (
		reset, clock : in std_logic;
		memdata_in : in std_logic_vector( 7 downto 0);
		memdata_out : out std_logic_vector ( 7 downto 0);
		address_out : out std_logic_vector( 7 downto 0);
		fetch_store : out std_logic;
		topout : out std_logic_vector( 7 downto 0);
		pcout : out std_logic_vector( 7 downto 0)
	);
end;

architecture structure of stack_processor is

	constant word_width : positive := 8;
	subtype Cell is std_logic_vector( word_width-1 downto 0);

  signal P, PP, R, M, T, F, U, D : Cell;  -- data paths
  signal nz : std_logic; -- not zero signal from top  to instruction
  signal mcontrol, pcontrol, tcontrol : std_logic_vector( 1 downto 0 );
  signal rcontrol, dcontrol : std_logic_vector( 2 downto 0 );
  signal fcontrol : std_logic_vector( 2 downto 0 );
  signal sp_opcode : Opcode;

  component instruction
    port ( iin   : in Cell;
           reset, nz, clock : in std_logic;
           iout  : out Opcode );
  end component;
  for all : instruction use entity work.instruction(behaviour2);

  component FunctionUnit
  	generic ( width : positive :=4 );
  	port ( 	selfunction :  in std_logic_vector( 2 downto 0 );
         	in1, in2    :  in std_logic_vector( width-1 downto 0 );
    		out1, out2  : out std_logic_vector( width-1 downto 0 )
		 );
  end component;
  for all : FunctionUnit use entity work.FunctionUnit(behaviour);

  component ProgramCounter
    port ( ain1, ain2 : in Cell;
           clock,reset: in std_logic;
           control    : in std_logic_vector( 1 downto 0);
           aoutport: out Cell);
  end component;
  for all : ProgramCounter use entity work.ProgramCounter(behaviour);

  component Stack

	generic ( width : positive := 8;
            widthad : positive := 3;
			depth : positive := 8 );

    port ( clock, push, pop, d1_d2 : in std_logic;
           din1, din2              : in Cell;
			reset					: in std_logic;
           dout                : out Cell);
  end component;
  for all : Stack use entity work.Stack(behaviour);

  component Top
	generic ( width : positive := 4 ); 
  	port ( din1, din2          : in std_logic_vector(width-1 downto 0);
         d1_d2, clock, store : in std_logic;
         notzero             : out std_logic;
         dout                : out std_logic_vector(width-1 downto 0));
  end component;
  for all : Top use entity work.Top(behaviour2);

begin

	pcout <= P;
	topout <= T;

     rcontrol(0)   <=       sp_opcode(0)  ;
     rcontrol(1)   <=       sp_opcode(1)  ;
     rcontrol(2)   <=       sp_opcode(2)  ;
     dcontrol(0)   <=       sp_opcode(3)   ;
     dcontrol(1)   <=       sp_opcode(4)   ;
     dcontrol(2)   <=       sp_opcode(5)   ;
     tcontrol(0)   <=       sp_opcode(6)  ;
     tcontrol(1)    <=      sp_opcode(7)   ;
     mcontrol(0)    <=      sp_opcode(8) ;
     mcontrol(1)    <=      sp_opcode(9)  ;
     pcontrol(0)   <=       sp_opcode(10) ;
     pcontrol(1)   <=       sp_opcode(11) ;
     fcontrol(0)    <=      sp_opcode(12) ;
     fcontrol(1)    <=      sp_opcode(13) ;
     fcontrol(2)   <=       sp_opcode(14) ;

	-- assign memory signals --
	fetch_store <= mcontrol(1);
	M <= memdata_in;
	memdata_out <= D;
	
	-- Memory --
	memory_address_select : process( T, P, mcontrol(0)) is
	begin
		-- mcontrol(0) = a1_a2 --
		if mcontrol(0)='0' then
			address_out <= P;	-- PC select
		else
			address_out <= T;	-- TOP select
		end if;
	end process;



  IR : instruction -- instruction register 
    port map ( iin => M, reset => reset, nz => nz, clock => clock,
               iout  => sp_opcode );
               
  DS : Stack  -- data stack
	generic map ( width => word_width, widthad => data_stack_index_width, 
					depth => data_stack_index_depth)
    port map ( reset => reset, clock => clock, push => dcontrol(2), pop => dcontrol(1),
               d1_d2 => dcontrol(0), din1 => U, din2 => M, dout => D );

  RS : Stack  -- return stack
	generic map ( width => word_width, widthad => return_stack_index_width, 
					depth => return_stack_index_depth)
    port map ( reset => reset, clock => clock, push => rcontrol(2), pop => rcontrol(1),
               d1_d2 => rcontrol(0), din1 => T, din2 => P, dout => R );

  PC : ProgramCounter
    port map ( ain1 => M, ain2 => R, clock => clock, reset => reset,
               control => pcontrol, aoutport => P );

  FU : FunctionUnit
	generic map ( width => Cell'length )
    port map ( selfunction => fcontrol,
               --leftin => lio, rightin => rio,
               --leftout => open, rightout => open,
               --leftinout => open, rightinout => open,
               in1 => T, in2 => D,
               out1 => F, out2 => U );
 
  TP : Top  -- top register
	generic map ( width => Cell'length )
    port map ( din1 => F, din2 => R, d1_d2 => tcontrol(0), clock => clock,
               store => tcontrol(1), notzero => nz, dout => T );


end structure;
